@@ -3,6 +3,7 @@ from nrgten.massdef import MassDef, massdefs_dicts_are_same
 import numpy as np
 import os
 import abc
+import sys
 
 
 class ENM(metaclass=abc.ABCMeta):
@@ -32,7 +33,8 @@ class ENM(metaclass=abc.ABCMeta):
         entropy (float): The computed vibrational entropy.
     """
     def __init__(self, pdb_file, added_atypes=None, added_massdef=None, atypes_list=None, massdef_list=None,
-                 verbose=False, solve=False, use_pickle=False, ignore_hetatms=False, solve_mol=True, one_mass=False):
+                 verbose=False, solve=False, use_pickle=False, ignore_hetatms=False, solve_mol=True, one_mass=False,
+                 teruel2021_legacy=False):
         """Constructor for the ENM abstract base class.
 
         Args:
@@ -52,6 +54,8 @@ class ENM(metaclass=abc.ABCMeta):
                                         of the residues will be inferred.
             one_mass (bool, optional): If True, nucleic acids will be built using only one mass per nucleotide instead
                                        of 3.
+            teruel2021_legacy (bool, optional): If True, the parameters will perfectly reproduce the vibrational entropy
+                                                calculations in the Teruel et al. 2021 PLoS Computational Biology paper
         """
         self.dirpath = os.path.dirname(os.path.abspath(__file__))
         assert len(self.dirpath) > 0
@@ -80,6 +84,7 @@ class ENM(metaclass=abc.ABCMeta):
         self.massdefs = self._parse_massdefs(atypes_list, massdef_list)
         self._print_verbose(self.massdefs)
         self.pdb_file = pdb_file
+        self.teruel2021_legacy = teruel2021_legacy
         self.mols = get_macromol_list(self.pdb_file, self.atypes_dict, self.massdefs, ignore_hetatms=ignore_hetatms)
         self.mol = self.mols[0]
         self.not_there = set()  # used to add resi-atom type pairs that are undefined
@@ -263,7 +268,7 @@ class ENM(metaclass=abc.ABCMeta):
         else:
             return bfacts
 
-    def compute_bfactors_boltzmann(self, beta=None, factor=1, filter=None, modes_proportion=1):
+    def compute_bfactors_boltzmann(self, beta=None, factor=1, filter=None, modes_proportion=1, use_xyz=False):
         assert self.eigvals is not None
         assert self.eigvecs is not None
         n = int(len(self.eigvecs) / 3)
@@ -272,22 +277,33 @@ class ENM(metaclass=abc.ABCMeta):
         if beta is None:
             beta = 1.55 * 10 ** -13  # beta is h/kT, this value is for T = 310 K
         beta *= factor
-        e = 2.718281828459045
-        pi = 3.1415926535897932384626433832795028841971693993751
         entros = [None, None, None, None, None, None]
         for j in range(6, end_j):
-            vj = (self.eigvals[j] ** 0.5) / (2 * pi)
+            vj = (self.eigvals[j] ** 0.5) / (2 * np.pi)
             x = vj * beta
-            entros.append(x / (e ** x - 1) - np.log(1 - e ** (-1 * x)))
+            entros.append(x / (np.e ** x - 1) - np.log(1 - np.e ** (-1 * x)))
         for i in range(n):
-            bfact = 0
-            for j in range(6, end_j):  # skips 1st 6 rotation-translation motions
-                temp = 0
+            if use_xyz:
+                bfacts_xyz = [0, 0, 0]
+                for j in range(6, end_j):  # skips 1st 6 rotation-translation motions
+                    for k in range(3):
+                        bfacts_xyz[k] += (self.eigvecs[j][3 * i + k] ** 2) * entros[j]
                 for k in range(3):
-                    temp += self.eigvecs[j][3 * i + k] ** 2
-                bfact += temp * entros[j]
-            bfacts.append(bfact * 1000)
-        filtered_bfacts = self.filter_bfactors(bfacts, filter)
+                    bfacts.append(bfacts_xyz[k] * 1000)
+            else:
+                bfact = 0
+                for j in range(6, end_j):  # skips 1st 6 rotation-translation motions
+                    temp = 0
+                    for k in range(3):
+                        temp += self.eigvecs[j][3 * i + k] ** 2
+                    bfact += temp * entros[j]
+                bfacts.append(bfact * 1000)
+        if filter is not None and use_xyz:
+            raise ValueError("Filtering not yet supported with xyz signatures")
+        if use_xyz:
+            filtered_bfacts = bfacts
+        else:
+            filtered_bfacts = self.filter_bfactors(bfacts, filter)
         self.bfacts = filtered_bfacts
         return filtered_bfacts
 
@@ -310,6 +326,8 @@ class ENM(metaclass=abc.ABCMeta):
         if beta is None:
             beta = 1.55 * 10 ** -13  # beta is h/kT, this value is for T = 310 K
         beta *= factor
+        if self.teruel2021_legacy:
+            beta *= 2*np.pi
         e = 2.718281828459045
         pi = 3.1415926535897932384626433832795028841971693993751
         entro = 0
@@ -319,6 +337,8 @@ class ENM(metaclass=abc.ABCMeta):
             entro += x / (e ** x - 1) - np.log(1 - e ** (-1 * x))
         return entro
 
+
+
     def compute_vib_enthalpy(self, beta=None, factor=1):
         """Vibrational enthalpy from the eigenfrequencies (without rigid-rotor approximation).
 
@@ -400,7 +420,8 @@ class ENM(metaclass=abc.ABCMeta):
                     current_j += 1
         return filtered_eigvecs[start:start+n_vecs]
 
-    def build_conf_ensemble(self, modes_list, filename, step=0.5, max_displacement=2.0, max_conformations=10000):
+    def build_conf_ensemble(self, modes_list, filename, step=0.5, max_displacement=2.0, max_conformations=10000,
+                            use_motion=False):
         """Creates a conformational ensemble as a multi-model PDB file.
 
         The idea is to make every combination of the selected modes at every given rmsd step, up to a total deformation
@@ -439,17 +460,80 @@ class ENM(metaclass=abc.ABCMeta):
                 raise ValueError(
                     "build_conf_ensemble was run with mode index {0} in the modes_list, which is a trivial " +
                     "(rotational/translational) motion.".format(mode_index))
-            eigvecs_list.append(np.copy(self.eigvecs[mode_index + 5]))
+            eigvecs_list.append(np.copy(self.eigvecs[mode_index - 1]))
             modermsd = self._rmsd_of_3n_vector(eigvecs_list[-1])
             eigvecs_list[-1] /= modermsd
         with open(filename, "w") as fh:
             fh.write("REMARK Conformational ensemble written by nrgten.enm.ENM.build_conf_ensemble()\n" +
                      "REMARK from the NRGTEN Python package, copyright Najmanovich Research Group 2020\n" +
                      "REMARK This ensemble contains {0} conformations\n".format(n_conf))
-            for i in range(n_conf):
+            conf_seq = range(n_conf)
+            if use_motion:
+                if len(modes_list) > 1:
+                    raise ValueError("Cannot use 'motion' style conformational ensemble with more than one mode")
+                conf_seq = []
+                for breaks in [[int((n_conf-1)/2), n_conf, 1], [n_conf-1, int((n_conf-1)/2)-1, -1],
+                               [int((n_conf-1)/2)-1, -1, -1], [0, int((n_conf-1)/2), 1]]:
+                    for i in range(*breaks):
+                        conf_seq.append(i)
+            for i in conf_seq:
                 self._write_one_point(eigvecs_list, i, grid_side, step, fh)
             fh.write("END\n")
 
+    def build_conf_ensemble_with_nrg(self, modes_list, filename_ensemble, filename_energies, base_energy, beta,
+                                     eigval_scaling=0.1, step=0.5, max_displacement=2.0, max_conformations=10000,
+                                     use_motion=False):
+        assert isinstance(modes_list, list)
+        eigvals_list = np.array([self.eigvals[x-1]*eigval_scaling for x in modes_list])
+
+        grid_side = 1 + (2 * max_displacement / step)  # length of one side of the grid
+        if abs((grid_side - 0.999999999999) % 2) > 0.000001:  # make sure that small floating point errors are ignored
+            raise ValueError(
+                "build_conf_ensemble was executed with step={0} and max_displacement={1}. max_displacement " +
+                "has to be a multiple of step.".format(step, max_displacement))
+        grid_side = int(round(grid_side))
+        grid_side = int(grid_side)
+        n_conf = int(grid_side ** len(modes_list))  # total number of points, or conformations, in the grid
+        if n_conf > max_conformations:
+            raise ValueError("build_conf_ensemble was executed with parameters specifying {0} conformations with " +
+                             "max_conformations set at {1}. If you really want that many conformations, set " +
+                             "max_conformations higher.".format(n_conf, max_conformations))
+        eigvecs_list = []
+        for mode_index in modes_list:
+            if mode_index < 7:
+                raise ValueError(
+                    "build_conf_ensemble was run with mode index {0} in the modes_list, which is a trivial " +
+                    "(rotational/translational) motion.".format(mode_index))
+            eigvecs_list.append(np.copy(self.eigvecs[mode_index - 1]))
+            modermsd = self._rmsd_of_3n_vector(eigvecs_list[-1])
+            eigvecs_list[-1] /= modermsd
+        energies = np.zeros(n_conf)
+        print(eigvals_list)
+        with open(filename_ensemble, "w") as fh:
+            fh.write("REMARK Conformational ensemble written by nrgten.enm.ENM.build_conf_ensemble()\n" +
+                     "REMARK from the NRGTEN Python package, copyright Najmanovich Research Group 2020\n" +
+                     "REMARK This ensemble contains {0} conformations\n".format(n_conf))
+            conf_seq = range(n_conf)
+            if use_motion:
+                if len(modes_list) > 1:
+                    raise ValueError("Cannot use 'motion' style conformational ensemble with more than one mode")
+                conf_seq = []
+                for breaks in [[int((n_conf - 1) / 2), n_conf, 1], [n_conf - 1, int((n_conf - 1) / 2) - 1, -1],
+                               [int((n_conf - 1) / 2) - 1, -1, -1], [0, int((n_conf - 1) / 2), 1]]:
+                    for i in range(*breaks):
+                        conf_seq.append(i)
+            for i in conf_seq:
+                energies[i] = self._write_one_point_nrg(eigvecs_list, eigvals_list, i, grid_side, step, fh)
+            fh.write("END\n")
+        energies += base_energy
+        probabilities = np.e**(-beta*energies)
+        probabilities /= np.sum(probabilities)
+        with open(filename_energies, "w") as f:
+            f.write("model energy beta probability\n")
+            for i in range(len(energies)):
+                f.write("{} {} {} {}\n".format(i+1, energies[i], beta, probabilities[i]))
+
+
     def _write_one_point(self, eigvecs_list, conf_n, grid_side, step, fh):
         """Helper function for build_conf_ensemble. conf_n is the conformation number from 0 to n-1.
 
@@ -470,6 +554,30 @@ class ENM(metaclass=abc.ABCMeta):
         self._write_model(int(model_n + 1), fh)
         self._translate_3n_vector(-t_vect)
 
+    def _write_one_point_nrg(self, eigvecs_list, spring_constants, conf_n, grid_side, step, fh):
+        """Helper function for build_conf_ensemble. conf_n is the conformation number from 0 to n-1.
+
+        Computes contributions from every mode automatically, translates the enm coordinates, and resets it back to
+        original after having written one model to the filehandle fh.
+        """
+        model_n = conf_n
+        nsteps_list = []
+        for i in range(len(eigvecs_list)):
+            nsteps = conf_n % grid_side
+            conf_n -= nsteps
+            conf_n /= grid_side
+            nsteps_list.append(nsteps - (grid_side - 1) / 2)
+        t_vect = np.zeros(len(eigvecs_list[0]))
+        nrg = 0
+        for vec, nsteps, k in zip(eigvecs_list, nsteps_list, spring_constants):
+            t_vect += vec * nsteps * step
+            nrg += k * (nsteps*step)**2
+        self._translate_3n_vector(t_vect)
+        self._write_model(int(model_n + 1), fh)
+        self._translate_3n_vector(-t_vect)
+        return nrg
+
+
     def _rmsd_of_3n_vector(self, vec):
         dists = np.zeros((int(len(vec) / 3)))
         for i in range(0, len(vec), 3):
@@ -632,11 +740,15 @@ class ENM(metaclass=abc.ABCMeta):
             for label, bfact in zip(labels, bfacts):
                 f.write("{0}\t{1}\n".format(label, bfact / divider))
 
-    def get_mass_labels(self):
+    def get_mass_labels(self, use_xyz=False):
         masses = self.mol.masses
         labels = []
         for mass in masses:
-            labels.append(mass[4])
+            if use_xyz:
+                for axis in ['X', 'Y', 'Z']:
+                    labels.append("{}|{}".format(mass[4], axis))
+            else:
+                labels.append(mass[4])
         return labels
 
     def _write_to_file(self, filename):
@@ -680,13 +792,98 @@ class ENM(metaclass=abc.ABCMeta):
             raise ValueError("Unable to produce {0} orthonormalized vectors for pdb file {1}".format(target_n, self.pdb_file))
         return np.array(basis)
 
+    def _get_weighted_surf_atomnum(self, sd, num_a, num_b, resi_a, resi_b, anums_dict):
+        try:
+            i = self.atypes_dict[resi_a][anums_dict[num_a].name]
+        except KeyError:
+            self.not_there.add((resi_a, anums_dict[num_a]))
+            return 0
+        try:
+            j = self.atypes_dict[resi_b][anums_dict[num_b].name]
+        except KeyError:
+            self.not_there.add((resi_b, anums_dict[num_b]))
+            return 0
+        return sd[num_a][num_b] * self.inter_mat[i][j]
+
+    def _get_total_surf_atomnum(self, sd, anums_list, i, j, resi_a, resi_b, anums_dict):
+        if resi_a == "LIG" and resi_b == "ASP":
+            dum = 0
+        surf_tot = 0
+        for num_a in anums_list[i]:
+            for num_b in anums_list[j]:
+                try:
+                    if num_b in sd[num_a]:
+                        surf_tot += self._get_weighted_surf_atomnum(sd, num_a, num_b, resi_a, resi_b, anums_dict)
+                    if num_a in sd[num_b]:
+                        surf_tot += self._get_weighted_surf_atomnum(sd, num_b, num_a, resi_b, resi_a, anums_dict)
+                except:
+                    print("Error in surface computation, residues {0} and {1}, file {2}".format(resi_a, resi_b,
+                                                                                                self.mol.pdb_file))
+                    raise
+        return surf_tot
+
+    def _compute_Bij(self, masses, resis):
+        """ Computes the Bij term in the ENCoM potential, which is the
+            modulating factor of the long-range interaction (V4) term calculated
+            according to atomic surface complementarity.
+        """
+        sd = self.mol.get_surface_dict()
+        n = len(masses)
+        bij = np.zeros((n, n))
+        atomnums_list = self._make_atomnums_list(masses, resis)
+        atomnums_dict = self._make_atomnums_dict(resis)
+        self._print_verbose(self.atypes_dict)
+        for i in range(n):
+            ma = masses[i]  # mass_a
+            ka = ma[4]  # key_a
+            resiname_a = ka.split('|')[0].split('.')[0]
+            for j in range(i + 1, n):
+                mb = masses[j]
+                kb = mb[4]
+                resiname_b = kb.split('|')[0].split('.')[0]
+                try:
+                    surf_tot = self._get_total_surf_atomnum(sd, atomnums_list, i, j, resiname_a, resiname_b, atomnums_dict)
+                except:
+                    e = sys.exc_info()[0]
+                    print("Error in surface computation, file {0}".format(self.mol.pdb_file))
+                    raise e
+                bij[i][j] = surf_tot
+                bij[j][i] = surf_tot
+        self._print_verbose(self.not_there)
+        return bij
+
+    def _make_atomnums_list(self, masses, resis):
+        """ Makes a list which corresponds in index with the masses received and
+            which contains lists of atom numbers for every mass.
+        """
+        assert len(masses) == len(resis)
+        n = len(masses)
+        atomnums_list = []
+        for i in range(n):
+            resi = resis[masses[i][4]]
+            anums = []
+            for a in resi:
+                anums.append(resi[a].num)
+            atomnums_list.append(anums)
+        return atomnums_list
+
+    def _make_atomnums_dict(self, resis):
+        """ Makes a dictionary associating each atom number to an Atom object.
+        """
+        numdict = dict()
+        for r in resis:
+            for a in resis[r]:
+                numdict[resis[r][a].num] = resis[r][a]
+        return numdict
+
+
 
 # End of ENM class, start of utility methods ###########################################################################
 
 
 def generate_dynasigs_df(filenames, outname, id_func=None, beta_values=None, models=None, models_labels=None,
-                         additional_info_dict=None, add_info_labels=None,
-                         added_atypes_list=None, added_massdef_list=None, get_cf_from_file=False):
+                         additional_info_dict=None, add_info_labels=None, added_atypes_list=None,
+                         added_massdef_list=None, get_cf_from_file=False, use_xyz=False):
     if models is None:
         from nrgten.encom import ENCoM
         models = [ENCoM]
@@ -706,16 +903,16 @@ def generate_dynasigs_df(filenames, outname, id_func=None, beta_values=None, mod
     vib_entro_data = []
     mod_labels_data = []
     cfs = []
-    for i, filename in enumerate(filenames):
-        filename_id = id_func(filename)
-        for mod, mod_lab in zip(models, models_labels):
+    for mod, mod_lab in zip(models, models_labels):
+        for i, filename in enumerate(filenames):
+            filename_id = id_func(filename)
             if added_atypes_list is None:
                 enm = mod(filename)
             else:
                 enm = mod(filename, added_atypes=added_atypes_list[i], added_massdef=added_massdef_list[i])
-            masslabels = [x.split('.')[-1] for x in enm.get_mass_labels()]  # only the mass name to allow for mutations
+            masslabels = [x.split('.')[-1] for x in enm.get_mass_labels(use_xyz=use_xyz)]  # only the mass name to allow for mutations
             for beta_val in beta_values:
-                dynasigs_data.append(enm.compute_bfactors_boltzmann(beta=beta_val))
+                dynasigs_data.append(enm.compute_bfactors_boltzmann(beta=beta_val, use_xyz=use_xyz))
                 vib_entro_data.append(enm.compute_vib_entropy(beta=beta_val))
                 if dynasigs_masslabels is None:
                     dynasigs_masslabels = masslabels
@@ -752,7 +949,7 @@ def generate_dynasigs_df(filenames, outname, id_func=None, beta_values=None, mod
             current_str = current_str + " {}".format(cfs[i])
         if additional_info_dict is not None:
             current_str = current_str + " " + " ".join(
-                [additional_info_dict[ids_data[i]][label] for label in add_info_labels])
+                [str(additional_info_dict[ids_data[i]][label]) for label in add_info_labels])
         current_str = current_str + " " + " ".join([str(x) for x in dynasigs_data[i]])
         df_strs.append(current_str)
     with open(outname, "w") as f:
